# Copyright (c) 2003 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Weaver module.
The weaver defines how to (des)apply aspects on classes
or instances
"""

__revision__ = '$Id: weaver.py,v 1.18 2003/06/16 15:23:39 adim Exp $'

import new
import types

from logilab.aspects.singleton import Singleton
from logilab.aspects._exceptions import AspectFailure, AlreadyAspectedError, \
     ClassNotWeaved

import sys, traceback
# from aspects.utils import extract_traceback



def make_wrapped_method(aspect):
    """Wraps a class method, and returns the wrapped method
    aspect : the aspect used to wrap the method
    returns the wrapped method
    """
    # we only want the func (not a (un)bound method)
    # func = ori.im_func
    # Define the wrapped method

    def wrapped(self, *args, **kwargs):
        """The wrapped method
        raises an AspectFailure if aspect fails
        """
        exec_excpt = None # The exception raised by the wrapped method
        tcbk = None       # The exception's traceback
        
        failure = None    # The aspect failure
        # bef is a shortcut for aspect.before => it's a bound method
        # (the real first arg (the aspect instance) is hidden here)
        # self represents the object weaved instance
        aspect.before(self, *args, **kwargs)

        # Catch the exception
        try:
            ret_v = aspect.around(self, *args, **kwargs)
        except:
            exec_excpt, tcbk = sys.exc_info()[1:]
            # Skip the first element of the traceback which refers to
            # this method ('wrapped')
            tcbk_list = traceback.format_tb(tcbk)[1:]
            # Append traceback to exception args
            exec_excpt.args += (''.join(tcbk_list),)
            # According to Python Doc, not deleting tcbk would cause a ref cycle
            del tcbk
            ret_v = None

        try:
            aspect.after(self, ret_v, exec_excpt, *args, **kwargs)
        except AspectFailure, failure:
            failure.set_method_exception(exec_excpt)
            raise failure

        # If no exception, then return ret_v
        if not exec_excpt:
            return ret_v
        # Else, re-raise exception
        else:
            raise exec_excpt

    # Return the wrapped method
    return wrapped



wrapped_def = """def wrapped_func(self, *args, **kwargs):
    
    exec_excpt = None # The exception raised by the wrapped method
    failure = None    # The aspect failure
    # bef is a shortcut for aspect.before => it's a bound method
    # (the real first arg (the aspect instance) is hidden here)
    # self represents the object weaved instance
    try:
        bef(self, *args, **kwargs)
    except AspectFailure:
        raise

    # Cattch the exception
    try:
        ret_v = func(self, *args, **kwargs)
    except Exception, excpt:
        exec_excpt = excpt
        ret_v = None

    try:
        aft(self, ret_v, exec_excpt, *args, **kwargs)
    except AspectFailure, failure:
        failure.set_method_exception(exec_excpt)
        raise failure

    # If no exception, then return ret_v
    if not exec_excpt:
        return ret_v
    # Else, re-raise exception
    else:
        raise exec_excpt

"""


class Weaver(Singleton):
    """The Weaver class weaves aspects in classes
    A Weaver object is used to change the code of a class
    in order to use an aspect.
    
    An example of woven_dict :
       {'AClass': {'method_f': {'original' : ori_method,
                                'aspects' : [Aspect1,Aspect2]}},
        'AnotherClass': {'method_h': ... }
        }
    """
    
    def __init__(self):
        """Initializes the woven dict
        """
        self.__woven_dict = {}


    def weave_module(self, mod, aspect_class, *args):
        """Will weave the aspect in all classes and instances
        found in module mod.
        """

        for obj_name in dir(mod):
            if not obj_name.startswith('_'):
                obj = getattr(mod, obj_name)
                obj_type = type(obj)
                if obj_type == types.ClassType or \
                   obj_type == types.InstanceType or \
                   obj_type == types.TypeType:
                    self.weave_methods(obj, aspect_class, *args)
        

    def unweave_module(self, mod, aspect_class, *args):
        """Will weave the aspect in all classes and instances
        found in module mod.
        """

        for obj_name in dir(mod):
            if not obj_name.startswith('_'):
                obj = getattr(mod, obj_name)
                obj_type = type(obj)
                if obj_type == types.ClassType or \
                   obj_type == types.InstanceType or \
                   obj_type == types.TypeType:
                    self.unweave_methods(obj, aspect_class, *args)


    def weave_methods(self, obj, aspect_class, *args):
        """ Will weave the aspect code in all the methods of obj
        obj : the class or class instance to weave
        aspect_class : the aspect's class we want to apply
        The varargs list is used in the aspect's class __init__ method
        """

        klass = obj
        if type(obj) == types.InstanceType:
            klass = obj.__class__
            # if object's class has already been aspected, then don't rewrap its
            # method, just add obj as a key in the dict 
            if self.obj_class_is_aspected(obj, aspect_class):
                print "Obj's class has already been aspected with ",aspect_class
                return
            
        method_list  = [item[0] for item in klass.__dict__.items()
                        if callable(item[1])]
        
        if obj not in self.__woven_dict:
            self.__woven_dict[obj] = {}

        for met_name in method_list:
            if not met_name.startswith("__"):
                aspect = aspect_class(obj, met_name, *args)
                self._weave_method(obj, met_name, aspect)

        

    def unweave_methods(self, obj, aspect_class = None):
        """ Will unweave the 'aspect' code in all the class 'cls'
        obj : the class or class instance to unweave
        aspect_class : the aspect's class we want to remove. If aspect_class is
        None, then remove it will unweave all aspects weaved on the class.
        """
        try:
            methods = self.__woven_dict[obj] # methods is a dict
        except KeyError:
            print "%s was not weaved"%obj
        else:
            for method_name, method_info in methods.items():
                # used to know if there are still aspects in the class after
                # we have unweaved methods
                aspects_found = 0

                aspects = method_info['aspects']
                # Make a list of all remaining aspects
                if aspect_class is None:
                    # None <=> Remove ALL aspects
                    remaining_aspects = []
                else:
                    remaining_aspects = [aspect for aspect in aspects
                                         if aspect.__class__ != aspect_class]
                    method_info['aspects'] = remaining_aspects
                
                aspects_found += len(remaining_aspects)
                # Retreive the base method (with no wrap at all)
                base_method = method_info['original']

                setattr(obj, method_name, base_method)
                # The new method is the base method wrapped by all
                # remaining aspects
                for aspect in remaining_aspects:
                    self._weave_method(obj, method_name, aspect)

            # Update woven dict if necessary
            if aspect_class is None or not aspects_found:
                del self.__woven_dict[obj]
        

    def _weave_method(self, obj, met_name, aspect):
        """ Will weave the 'aspect' code for a specific method 'm'
        obj : the class or the class instance to weave
        met_name : the method name
        aspect_class : the aspect class
        """

        try:
            methods = self.__woven_dict[obj] # method is a dict
            if met_name not in methods:
                self.__woven_dict[obj][met_name] = {
                    'original' : getattr(obj, met_name),
                    'aspects' : []
                    }
                
            # add method to woven_dict, and replace it by a wrapped method
            # in its original class
            try:
                self._add_aspect(obj, met_name, aspect)
            except AlreadyAspectedError, excpt:
                print excpt
                return
            
            # self.__woven_dict[obj][met_name]['aspects'].append(aspect)
            original_method = getattr(obj, met_name)
##             wrapped_func = make_wrapped_method(original_method,
##                                               aspect.before,
##                                               aspect.after)

            wrapped_func = make_wrapped_method(aspect)
            # This is important if we want the wrapped_func to behave
            # totally like the original one
            wrapped_func.__doc__ = original_method.__doc__
            
            # if it's a bound method (in all cases : obj <=> im_self)
            if original_method.im_self is not None:
                met = new.instancemethod(wrapped_func, obj, obj.__class__)
            else:
                met = new.instancemethod(wrapped_func, None, obj)
            setattr(obj, met_name, met)
            
        except KeyError:
            raise ClassNotWeaved, "Class %s was not weaved" % obj
        

        

    def _unweave_method(self, cls, met_name, aspect_class):
        """ Will unweave the 'aspect' code for the method 'm'
        cls : the class to unweave
        met_name : the method name
        aspect_class : the aspect_class that you want to remove
        """

        # Remove the method from the method_list, and then re-use
        # the original method
        try:
            methods = self.__woven_dict[cls] # methods is a dict !
            if met_name in methods:

                aspects = self.get_aspects(cls, met_name)
                # l should be a list of length 0 or 1 : we shouldn't have
                # several time the same aspect on the same method !
                matching_aspects = [aspect for aspect in aspects
                                    if aspect.__class__ == aspect_class]
                try:
                    self._remove_aspect(cls, met_name, matching_aspects[0])
                    # methods[met_name]['aspects'].remove(l[0])
                    setattr(cls, met_name,
                    matching_aspects[0].get_wrapped_method())
                except ValueError:
                    print "Aspect %r was not weaved in %s"%(aspect,met_name)
        except KeyError:
            raise ClassNotWeaved, "Class %s was not weaved" % obj


    def __str__(self):
        info  = []
        for cls, methods in self.__woven_dict.items():
            info.append('Class %s : '%cls.__name__)
            for met_name, aspects in methods.items():
                info.append('\tAspects weaved for method %s : '%met_name)
                for aspect in aspects:
                    info.append('\t\t - %s'%aspect)
        return '\n'.join(info)

    
    ## Weaver Helper Part ######################################################
    def get_weaved_classes(self):
        """Returns the list of the aspected classes
        """
        weaved_objects = self.__woven_dict.keys()
        return [weaved for weaved in weaved_objects
                if type(weaved) == types.ClassType]

    def get_weaved_instances(self):
        """Returns the list of the aspected instances
        """
        weaved_objects = self.__woven_dict.keys()
        return [weaved for weaved in weaved_objects
                if type(weaved) != types.ClassType]

    def get_weaved_methods(self, cls):
        """
        cls : the class to look at
        returns the weaved methods' names
        """
        try:
            return self.__woven_dict[cls].keys()
        except KeyError:
            raise ClassNotWeaved, "%s is not weaved"%cls
        

    def get_aspects(self, obj, met_name):
        """Returns the list of aspects weaved for a method
        obj : the class or class instance to look at
        met_name : the method's name
        """
        try:
            return self.__woven_dict[obj][met_name]['aspects']
        except KeyError:
            raise ClassNotWeaved, "%s.%s is not weaved "%(obj, met_name)


    def _add_aspect(self, obj, met_name, aspect):
        """Adds an aspect in the woven_dict
        obj : the class (or class instance)
        met_name : the method's name
        aspect : the aspect to add
        """
        aspect_list = self.__woven_dict[obj][met_name]['aspects']
        for added_aspect in aspect_list:
            if added_aspect.__class__ == aspect.__class__:
                raise AlreadyAspectedError, "%s is already aspected with %s"% \
                      (met_name, aspect.__class__.__name__)
            
        self.__woven_dict[obj][met_name]['aspects'].append(aspect)


    def _remove_aspect(self, obj, met_name, aspect):
        """Removes an aspect in the woven_dict
        obj : the class (or class instance)
        met_name : the method's name
        aspect : the aspect to remove

        Raises a ValueError if aspect is not in the list
        """
        self.__woven_dict[obj][met_name]['aspects'].remove(aspect)


    def obj_class_is_aspected(self, obj, aspect_class):
        """ Checks if 
        obj : the object to check
        aspect_class : the aspect_class

        returns True if aspect_class has already been woven in
        obj's class
        """
        if obj.__class__ in self.__woven_dict:
            class_info = self.__woven_dict[obj.__class__]
            aspect_classes = [aspect_class for aspect_class in class_info]
            if aspect_class in aspect_classes:
                # self.__woven_dict[obj] = dict(class_info)
                return True

        return False


# create global instance
weaver = Weaver()
